import os

from repo_smith.initialize_repo import initialize_repo


def test_reset_step():
    # only test hard reset (other valid modes can be assumed to work if hard works)
    repo_initializer = initialize_repo("tests/specs/reset_step/reset_step_hard.yml")
    with repo_initializer.initialize() as r:
        assert r.head.commit.message.strip() == "Initial commit"
        file_path = os.path.join(r.working_dir, "file1.txt")
        with open(file_path, "r") as f:
            content = f.read()
        assert "Initial content" in content
        assert "Modified content" not in content


def test_reset_step_files():
    repo_initializer = initialize_repo("tests/specs/reset_step/reset_step_files.yml")
    with repo_initializer.initialize() as r:
        staged_files = [d.a_path for d in r.index.diff("HEAD")]
        assert "file1.txt" not in staged_files
        assert "file2.txt" in staged_files



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class RevertOptions(TypedDict, total=False):
    no_commit: bool


REVERT_SPEC = CommandSpec().flag("no_commit", "--no-commit", default=False)



from typing import Dict, List, Optional

from git import Repo
from repo_smith.command_result import CommandResult, run


class Helper:
    def __init__(self, repo: Optional[Repo], verbose: bool) -> None:
        self.repo = repo
        self.verbose = verbose

    def run(
        self,
        command: List[str],
        env: Dict[str, str] = {},
        exit_on_error: bool = False,
    ) -> CommandResult:
        return run(command, self.verbose, env, exit_on_error)



import os

from shutil import copyfile
from git import Repo
from repo_smith.initialize_repo import initialize_repo
from tests.fixtures.git_fixtures import REMOTE_REPO_PATH, remote_repo


def test_revert_step_hash(remote_repo: Repo):
    (REMOTE_REPO_PATH / "dummy1.txt").write_text("first")
    remote_repo.index.add(["dummy1.txt"])
    remote_repo.index.commit("first commit")

    (REMOTE_REPO_PATH / "dummy2.txt").write_text("second")
    remote_repo.index.add(["dummy2.txt"])
    remote_repo.index.commit("second commit")

    (REMOTE_REPO_PATH / "dummy3.txt").write_text("third")
    remote_repo.index.add(["dummy3.txt"])
    remote_repo.index.commit("third commit")

    (REMOTE_REPO_PATH / "dummy4.txt").write_text("fourth")
    remote_repo.index.add(["dummy4.txt"])
    remote_repo.index.commit("fourth commit")

    full_hash = remote_repo.commit("HEAD~1").hexsha

    copyfile(
        "tests/specs/revert_step/revert_step_hash.yml",
        "tests/specs/revert_step/temp-1.yml"
        )

    with open("tests/specs/revert_step/temp-1.yml", "a") as f:
        f.write(full_hash)

    ir = initialize_repo("tests/specs/revert_step/temp-1.yml")
    with ir.initialize() as r:
        commits = list(r.iter_commits("main"))
        commit = commits[0]
        assert "Revert" in commit.message

    os.remove("tests/specs/revert_step/temp-1.yml")


def test_revert_step_short_hash(remote_repo: Repo):
    (REMOTE_REPO_PATH / "dummy1.txt").write_text("first")
    remote_repo.index.add(["dummy1.txt"])
    remote_repo.index.commit("first commit")

    (REMOTE_REPO_PATH / "dummy2.txt").write_text("second")
    remote_repo.index.add(["dummy2.txt"])
    remote_repo.index.commit("second commit")

    (REMOTE_REPO_PATH / "dummy3.txt").write_text("third")
    remote_repo.index.add(["dummy3.txt"])
    remote_repo.index.commit("third commit")

    (REMOTE_REPO_PATH / "dummy4.txt").write_text("fourth")
    remote_repo.index.add(["dummy4.txt"])
    remote_repo.index.commit("fourth commit")

    short_hash = remote_repo.commit("HEAD~1").hexsha[:7]

    copyfile(
        "tests/specs/revert_step/revert_step_short_hash.yml",
        "tests/specs/revert_step/temp-2.yml"
        )

    with open("tests/specs/revert_step/temp-2.yml", "a") as f:
        f.write(short_hash)

    ir = initialize_repo("tests/specs/revert_step/temp-2.yml")
    with ir.initialize() as r:
        commits = list(r.iter_commits("main"))
        commit = commits[0]
        assert "Revert" in commit.message

    os.remove("tests/specs/revert_step/temp-2.yml")


def test_revert_step_relative(remote_repo: Repo):
    (REMOTE_REPO_PATH / "dummy1.txt").write_text("first")
    remote_repo.index.add(["dummy1.txt"])
    remote_repo.index.commit("first commit")

    (REMOTE_REPO_PATH / "dummy2.txt").write_text("second")
    remote_repo.index.add(["dummy2.txt"])
    remote_repo.index.commit("second commit")

    (REMOTE_REPO_PATH / "dummy3.txt").write_text("third")
    remote_repo.index.add(["dummy3.txt"])
    remote_repo.index.commit("third commit")

    (REMOTE_REPO_PATH / "dummy4.txt").write_text("fourth")
    remote_repo.index.add(["dummy4.txt"])
    remote_repo.index.commit("fourth commit")

    ir = initialize_repo("tests/specs/revert_step/revert_step_relative.yml")
    with ir.initialize() as r:
        commits = list(r.iter_commits("main"))
        commit = commits[0]
        assert "Revert" in commit.message



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class ResetOptions(TypedDict, total=False):
    soft: bool
    mixed: bool
    hard: bool
    merge: bool
    keep: bool


RESET_SPEC = (
    CommandSpec()
    .flag("soft", "--soft", default=False)
    .flag("mixed", "--mixed", default=False)
    .flag("hard", "--hard", default=False)
    .flag("merge", "--merge", default=False)
    .flag("keep", "--keep", default=False)
)



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class InitOptions(TypedDict, total=False):
    bare: bool
    template: str
    branch: str
    initial_branch: str


INIT_SPEC = (
    CommandSpec()
    .flag("bare", "--bare", default=False)
    .opt("template", "--template")
    .opt("branch", "-b")
    .opt("initial_branch", "--initial-branch", default="main")
)



import pytest

from repo_smith.steps.revert_step import RevertStep


def test_add_step_parse_missing_files():
    with pytest.raises(ValueError, match='Missing "revision" field in revert step.'):
        RevertStep.parse("n", "d", "id", {})


def test_add_step_parse_empty_files():
    with pytest.raises(ValueError, match='Empty "revision" field in revert step.'):
        RevertStep.parse("n", "d", "id", {"revision": ""})


def test_revert_step_parse():
    step = RevertStep.parse("n", "d", "id", {"revision": "HEAD~4"})
    assert isinstance(step, RevertStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.revision == "HEAD~4"



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class MergeOptions(TypedDict, total=False):
    commit: bool
    no_commit: bool

    ff: bool
    no_ff: bool
    ff_only: bool

    squash: bool
    no_squash: bool

    message: str


MERGE_SPEC = (
    CommandSpec()
    .flag("commit", "--commit", default=False)
    .flag("no_commit", "--no-commit", default=False)
    .flag("ff", "--f", default=False)
    .flag("no_ff", "--no-ff", default=False)
    .flag("ff_only", "--ff-only", default=False)
    .flag("squash", "--squash", default=False)
    .flag("no_squash", "--no-squash", default=False)
    .opt("message", "-m")
)



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class BranchOptions(TypedDict, total=False):
    delete: bool
    force: bool
    move: bool
    copy: bool

    set_upstream_to: str
    unset_upstream: bool


BRANCH_SPEC = (
    CommandSpec()
    .flag("delete", "--delete", default=False)
    .flag("force", "--force", default=False)
    .flag("move", "--move", default=False)
    .flag("copy", "--copy", default=False)
    .opt("set_upstream_to", "--set-upstream-to")
    .flag("unset_upstream", "--unset-upstream", default=False)
)



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class FetchOptions(TypedDict, total=False):
    all: bool
    no_all: bool

    atomic: bool
    force: bool

    no_tags: bool
    tags: bool


FETCH_SPEC = (
    CommandSpec()
    .flag("all", "--all", default=False)
    .flag("no_all", "--no-all", default=False)
    .flag("atomic", "--atomic", default=False)
    .flag("force", "--force", default=False)
    .flag("no_tags", "--no-tags", default=False)
    .flag("tags", "--tags", default=False)
)



from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class RevertStep(Step):
    revision: str

    step_type: StepType = field(init=False, default=StepType.REVERT)

    def execute(self, repo: Repo) -> None:
        revert_args = [self.revision, "--no-edit"]

        repo.git.revert(*revert_args)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "revision" not in step:
            raise ValueError('Missing "revision" field in revert step.')

        if step["revision"] is None or step["revision"].strip() == "":
            raise ValueError('Empty "revision" field in revert step.')

        return cls(
            name=name,
            description=description,
            id=id,
            revision=step["revision"],
        )



import tempfile
from contextlib import contextmanager
from typing import Dict, Iterator, Optional, Self, Type, TypedDict, TypeVar, Unpack

from git.repo import Repo

from repo_smith.helpers.files_helper import FilesHelper
from repo_smith.helpers.git_helper.git_helper import GitHelper
from repo_smith.helpers.helper import Helper

T = TypeVar("T", bound="Helper")


class RepoSmith:
    def __init__(self, repo: Optional[Repo], verbose: bool) -> None:
        self.verbose = verbose
        self.__repo = repo
        self.files = FilesHelper(repo, verbose)
        self.git = GitHelper(repo, verbose)
        self.__additional_helpers: Dict[Type[Helper], Helper] = {}

    @property
    def repo(self) -> Repo:
        if self.__repo is None:
            raise ValueError("Repo is None, cannot access")
        return self.__repo

    def add_helper(self, cls: Type[T]) -> Self:
        self.__additional_helpers[cls] = cls(self.repo, self.verbose)
        return self

    def helper(self, cls: Type[T]) -> T:
        if cls not in self.__additional_helpers:
            raise ValueError(f"Additional helper of {cls.__name__} not found.")
        return self.__additional_helpers[cls]


class CreateRepoOptions(TypedDict, total=False):
    clone_from: str
    existing_path: str
    null_repo: bool


@contextmanager
def create_repo_smith(
    verbose: bool, **options: Unpack[CreateRepoOptions]
) -> Iterator[RepoSmith]:
    """Creates a RepoSmith instance over a given repository."""
    clone_from = options.get("clone_from")
    existing_path = options.get("existing_path")
    null_repo = options.get("null_repo", False)

    dir = tempfile.mkdtemp() if existing_path is None else existing_path

    if null_repo:
        repo = None
    elif clone_from:
        repo = Repo.clone_from(clone_from, dir)
    else:
        repo = Repo.init(dir, initial_branch="main")

    yield RepoSmith(repo, verbose)



import pytest

from repo_smith.steps.reset_step import ResetStep


def test_reset_step_parse_with_revision_and_mode():
    step = ResetStep.parse("n", "d", "id", {"revision": "HEAD~1", "mode": "mixed"})
    assert isinstance(step, ResetStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.revision == "HEAD~1"
    assert step.mode == "mixed"
    assert step.files is None


def test_reset_step_parse_with_files():
    step = ResetStep.parse(
        "n",
        "d",
        "id",
        {"revision": "HEAD", "mode": "mixed", "files": ["file1.txt", "file2.txt"]},
    )
    assert step.revision == "HEAD"
    assert step.mode == "mixed"
    assert step.files == ["file1.txt", "file2.txt"]


def test_reset_step_parse_missing_mode():
    with pytest.raises(ValueError, match='Missing "mode" field in reset step.'):
        ResetStep.parse("n", "d", "id", {"revision": "HEAD~1"})


def test_reset_step_parse_missing_revision():
    with pytest.raises(ValueError, match='Missing "revision" field in reset step.'):
        ResetStep.parse("n", "d", "id", {"mode": "hard"})


def test_reset_step_parse_empty_revision():
    with pytest.raises(ValueError, match='Empty "revision" field in reset step.'):
        ResetStep.parse("n", "d", "id", {"revision": "", "mode": "hard"})


def test_reset_step_parse_invalid_mode():
    with pytest.raises(
        ValueError,
        match="Invalid \"mode\" value. Must be one of: \\('soft', 'mixed', 'hard'\\).",
    ):
        ResetStep.parse("n", "d", "id", {"revision": "HEAD~1", "mode": "invalid"})


def test_reset_step_parse_empty_files_list():
    with pytest.raises(ValueError, match='Empty "files" list in reset step.'):
        ResetStep.parse(
            "n", "d", "id", {"revision": "HEAD", "mode": "mixed", "files": []}
        )


def test_reset_step_parse_files_with_soft_mode():
    with pytest.raises(
        ValueError,
        match='Cannot use "files" field with "soft" mode in reset step. Only "mixed" mode is allowed with files.',
    ):
        ResetStep.parse(
            "n", "d", "id", {"revision": "HEAD", "mode": "soft", "files": ["file.txt"]}
        )


def test_reset_step_parse_files_with_hard_mode():
    with pytest.raises(
        ValueError,
        match='Cannot use "files" field with "hard" mode in reset step. Only "mixed" mode is allowed with files.',
    ):
        ResetStep.parse(
            "n", "d", "id", {"revision": "HEAD", "mode": "hard", "files": ["file.txt"]}
        )



from dataclasses import dataclass, field
from typing import Any, List, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType

VALID_MODES = ("soft", "mixed", "hard")


@dataclass
class ResetStep(Step):
    revision: Optional[str]
    mode: str
    files: Optional[List[str]]

    step_type: StepType = field(init=False, default=StepType.RESET)

    def execute(self, repo: Repo) -> None:
        if self.files:
            repo.git.reset(self.revision, "--", *self.files)
        else:
            repo.git.reset(f"--{self.mode}", self.revision)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "mode" not in step:
            raise ValueError('Missing "mode" field in reset step.')

        if step["mode"] is None or step["mode"].strip().lower() not in VALID_MODES:
            raise ValueError(f'Invalid "mode" value. Must be one of: {VALID_MODES}.')

        if "revision" not in step:
            raise ValueError('Missing "revision" field in reset step.')

        if step["revision"] is None or step["revision"].strip() == "":
            raise ValueError('Empty "revision" field in reset step.')

        if "files" in step and step["files"] is not None:
            if step["files"] == []:
                raise ValueError('Empty "files" list in reset step.')

            if step["mode"] != "mixed":
                raise ValueError(
                    f'Cannot use "files" field with "{step["mode"]}" mode in reset step. Only "mixed" mode is allowed with files.'
                )

        return cls(
            name=name,
            description=description,
            id=id,
            revision=step["revision"],
            mode=step["mode"],
            files=step.get("files", None),
        )
